mod oop;
use std::{thread::{self, JoinHandle}, io::Write};
use roxmltree;
use crossbeam;

fn main() {
    let args: Vec<String> = std::env::args().collect();

    let mut file_names: Vec<String> = Vec::new();
    let mut handles:Vec<JoinHandle<()>> = Vec::new();
    let opt = roxmltree::ParsingOptions {
        allow_dtd: true,
        ..roxmltree::ParsingOptions::default()
    };

    for arg in args[1..].iter().cloned(){
        file_names.push(arg.to_string());
        handles.push(thread::spawn(move || {
            println!("{:?}", arg);
            if arg.to_string().contains(".xml") {
                let contents = std::fs::read_to_string(arg).expect("read_to_string didn't work");
                let doc = roxmltree::Document::parse_with_options(&contents, opt);
                //println!("{:?}", doc);
                match doc {
                    Ok(doc_good) => {
                        //output(parse(&doc_good));
                        //#[allow(unused_variables)]
                        let family = new_parse(&doc_good).unwrap();
                        let output_paths = [family.get_name() + "_objects.rs", family.get_name() + "_members.rs"];
                        match crossbeam::scope(|s| {
                            for output_path in output_paths {
                                s.spawn(|_| {
                                    write_file(output_gen(family.clone(), output_path))
                                });
                            }
                        }) {
                            Ok(no_panic) => println!("Subthreads didn't panic:  {:?}", no_panic),
                            Err(thread_panic) => println!("Subthread panic: {:?}", thread_panic)
                        }
                            
                            
                    },
                    Err(doc_bad) => {
                        println!("{:?}", doc_bad);
                        panic!();
                    }
                }
            }
        }));
    };

    for handle in handles {
        handle.join().expect("Couldn't close thread.");
    }
}
fn new_parse(doc:&roxmltree::Document) -> Result<oop::Family,core::fmt::Error>{
    let current = oop::Current::new();
    let mut wrapped_family:Option<oop::Family> = None;
    let mut messages: oop::Messages = oop::Messages::new(Vec::new(), Vec::new(), Vec::new(), Vec::new(), Vec::new());
    let root = doc.root_element();

    //println!("{:?}",doc.root_element());
    //println!("{:?}",doc.root_element().first_element_child());
    if let Some(_) = doc.root_element().first_element_child(){
        match root.tag_name().name() {
            "FAMILY" => {
                for attribute in root.attributes().filter(|attribute| { attribute.name() == "name" }){
                    println!("successful match");
                    wrapped_family = Some(oop::Family::new(attribute.value().to_string(), Vec::new(), Vec::new()));
                }
            },
            _ => {}
        }

        if let Some(ref mut family) = wrapped_family{
            root.children().for_each(|child| nested_parse(family, child, current.clone()));
        }
        root.children().filter(|outer_child| outer_child.tag_name().name() == "MESSAGES" )
            .for_each(|message_group| {
                message_group.children().filter(|inner_child| inner_child.tag_name().name() == "MSG")
                .for_each(|message| {
                    messages.push(message);
                });
        });

        if let Some(mut family) = wrapped_family{
            family.push_messages(messages);
            return Ok(family);
        }
    }
    return Err(core::fmt::Error);
}

fn nested_parse(input_family:&mut oop::Family, node: roxmltree::Node, mut current: oop::Current) {
    if node.is_text() || node.is_comment() { return; }
    //println!("{:?}",node);
    match node.tag_name().name() {
        "SPECIES_KEY" => {
            node.children().for_each(|child| nested_parse(input_family, child, current.clone()));
        },
        "GENUS" => {
            node.attributes()
                .filter(|attr| attr.name() != "comment")
                .for_each(|attr| {
                    if attr.name() == "name" { 
                        input_family.push_genus(oop::Genus::new(attr.value().to_string(), Vec::new()));
                    };
                });
            node.children().for_each(|child| nested_parse(input_family, child, current.clone()));
        },
        "SPECIES" => {
            input_family.get_genera().last_mut().unwrap()
                .push_species(oop::Species::new(
                    node.attribute("name").unwrap().to_string(),
                    node.attribute("defaultvalue").unwrap_or_default().to_string(),
                    Vec::new()
                ));
        },
        "DEFINITION" => {
            current.attr_family = node.attribute("family").unwrap().to_string();
            current.attr_genus = node.attribute("genus").unwrap().to_string();
            current.attr_species = node.attribute("species").unwrap().to_string();
            if current.attr_family != input_family.get_name() { 
                println!("Inconsistent family naming scheme. Check XML file.");
                panic!();
            }
            input_family.get_genera().as_mut_slice().iter_mut()
                .filter(|genus| genus.get_name().as_str() == current.attr_genus)
                .for_each(|genus| {
                    genus.get_species().as_mut_slice().iter_mut()
                        .filter(|species| species.get_name().as_str() == current.attr_species)
                        .for_each(|species|{
                            current.attr_object = node.attribute("class").unwrap().to_string();
                            species.push_object(oop::Object::new(
                                node.attribute("class").unwrap().to_string(),
                                node.attribute("abrv").unwrap().to_string(),
                                node.attribute("version").unwrap().to_string(),
                                Vec::new()
                            ));
                        });
                }
            );
            node.children().for_each(|child| nested_parse(input_family, child, current.clone()));
        },
        "MEMBERS" => {
            node.children().for_each(|child| nested_parse(input_family, child, current.clone()));
        },
        "MEMBER" => {
            input_family.get_genera().as_mut_slice().iter_mut()
                .filter(|genus| genus.get_name().as_str() == current.attr_genus)
                .for_each(|genus| {
                    genus.get_species().as_mut_slice().iter_mut()
                        .filter(|species| species.get_name().as_str() == current.attr_species)
                        .for_each(|species|{
                            current.attr_member = node.attribute("name").unwrap().to_string();
                            let size:Option<usize>;
                            match node.attribute("size"){
                                None => size = None,
                                Some(size_string) => {
                                    if size_string == "UNBOUNDED" {
                                        size = Some(usize::MAX);
                                    } else {
                                        size = Some(size_string.parse::<usize>().unwrap());
                                    }
                                },
                            }
                                species.get_objects().last_mut().unwrap().push_object_member(oop::ObjectMember::new(
                                    node.attribute("type").unwrap().to_string(),
                                    node.attribute("name").unwrap().to_string(),
                                    node.attribute("class").unwrap_or("").to_string(),
                                    node.attribute("minversion").unwrap().to_string(),
                                    node.attribute("maxversion").unwrap().to_string(),
                                    Vec::new(),
                                    node.attribute("format").unwrap_or("").to_string(),
                                    size
                                ));
                            
                        });
                }
            );
        },
        "ENUMERATION" => {
            current.attr_member = node.attribute("name").unwrap().to_string();
            node.children().for_each(|child| nested_parse(input_family, child, current.clone()));
        },
        "ENUM" => {
            input_family.get_genera().as_mut_slice().iter_mut()
                .filter(|genus| genus.get_name().as_str() == current.attr_genus)
                .for_each(|genus| {
                    genus.get_species().as_mut_slice().iter_mut()
                        .filter(|species| species.get_name().as_str() == current.attr_species)
                        .for_each(|species|{
                            species.get_objects().as_mut_slice().iter_mut().filter(|object| object.get_name().as_str() == current.attr_object)
                                .for_each(|object| {
                                    object.get_members().as_mut_slice().iter_mut().filter(|member| member.get_member_name().as_str() == current.attr_member)
                                        .for_each(|member| {
                                            member.get_enumerations().push(
                                                oop::MemberEnumeration::new(
                                                    node.attribute("name").unwrap().to_string(),
                                                    node.attribute("defaultvalue").unwrap_or("").to_string(),
                                                )
                                            );
                                        });
                            });
                        });
                }
            );
        },
        //These two are handled elsewhere
        "MESSAGES" => {},
        "MSG" => {},
        
        //catch-all
        _ => {}
    }
}

fn output_gen(family: oop::Family, output_path: String) -> (String, String){
    let mut write_string: String;

    match output_path.as_str() {
        o if o.contains("objects") => {
            write_string = format!(
                "{}{}{}{}{}", 
                "use crate::datatypes;\nuse super::{Object, ObjectSpecies, RawObject};\nuse derive_getters::Getters;\nuse derive_builder::Builder;\nuse downcast_rs::Downcast;\n\npub trait ",
                family.get_name(),
                ":std::fmt::Debug + Downcast{}\n\n#[derive(Clone,Debug)]\npub enum ",
                family.get_name(),
                "Species{\n"
            );
            for species_name in family.list_species() {
                write_string.push_str(format!("\t{}\n",species_name).as_str());
            }
            write_string.push_str(format!(
                "}}\nimpl ObjectSpecies for {}Species{{}}\n\n",
                family.get_name()
            ).as_str());
            let mut genus_name:String;
            for genus in family.clone().get_genera() {
                genus_name = genus.get_name();
                for species in genus.get_species() {
                    for object in species.get_objects() {
                        write_string.push_str(format!(
                            "impl {} for {}{{}}\n#[derive(Getters,Builder,Clone,Debug)]\npub struct {}{{\n", 
                            family.get_name(),
                            object.get_name(),
                            object.get_name(),
                        ).as_str());
                        let mut enums = Vec::new();
                        for member in object.get_members() {
                            match member.get_member_type().as_str(){
                                "OBJECT" => {
                                    write_string.push_str(format!(
                                        "\t{}:&'static dyn Object,\n", 
                                        member.get_member_name().to_lowercase(),
                                    ).as_str());
                                },
                                "TIME" => {
                                    write_string.push_str(format!(
                                        "\t{}:chrono::NaiveDateTime,\n",
                                        member.get_member_name().to_lowercase())
                                    .as_str());
                                },
                                "ansichar" | "widechar" => {
                                    write_string.push_str(format!(
                                        "\t{}:String,\n",
                                        member.get_member_name().to_lowercase())
                                    .as_str());
                                },
                                _ => {
                                    write_string.push_str(format!(
                                        "\t{}:", 
                                        member.get_member_name().to_lowercase(),
                                    ).as_str());
                                    if let Some(size_value) = member.get_size(){
                                        if size_value == usize::MAX {
                                            write_string.push_str("Vec<");
                                        } else {
                                            write_string.push_str("[");
                                        }
                                    }
                                    if member.get_member_type().contains("uint"){
                                        let mut fixed_type = member.get_member_type();
                                        fixed_type.retain(|char| char.is_digit(10) || char == 'u' );
                                        write_string.push_str(fixed_type.as_str());
                                    } else if member.get_member_type().contains("int"){
                                        let mut fixed_type = member.get_member_type();
                                        fixed_type.retain(|char| char.is_digit(10) || char == 'i' );
                                        write_string.push_str(fixed_type.as_str());
                                    } else if member.get_member_type().contains("bool"){
                                        write_string.push_str("bool");
                                    } else if member.get_member_type().contains("float"){
                                        write_string.push_str("f32");
                                    } else { 
                                        write_string.push_str(member.get_member_type().as_str()); 
                                    }

                                    if let Some(size_value) = member.get_size(){
                                        if size_value == usize::MAX {
                                            write_string.push_str(">,\n");
                                        } else {
                                            write_string.push_str(format!("; {},\n]",size_value).as_str());
                                        }
                                    } else {
                                            write_string.push_str(",\n");
                                    }
                                },
                            }
                            if !member.get_enumerations().is_empty() && member.get_enumerations().first().unwrap().get_name() != "" {
                                enums.push(member)
                            }
                        }
                        write_string.push_str("\t#[getter(skip)]\n\traw_object:RawObject,\n");
                        
                        if !enums.is_empty() {
                            for member in enums {
                                write_string.push_str(format!(
                                    "{}{}{}",
                                    "}\n\n#[derive(Clone,Copy,Debug,Default)]\npub enum ",
                                    member.get_member_name(),
                                    "{\n\t#[default]\n"  
                                ).as_str());
                                for e in member.get_enumerations() {
                                    write_string.push_str(format!(
                                        "\t{},\n",
                                        e.get_name(),
                                    ).as_str());
                                }
                            }
                        }
                        write_string.push_str(format!(
                            "{}{}{}{}{}{}{}{}{}{}{}{}{}",
                            "}\n\nimpl ",
                            object.get_name(),
                            " {\n\tpub fn new() -> Self { todo!() }\n}\nimpl Object for ",
                            object.get_name(),
                            "{\n\tfn get_family(&self) -> datatypes::Family { datatypes::Family::",
                            family.get_name(),
                            " }\n\tfn get_genus(&self) ->  datatypes::Genus { datatypes::Genus::",
                            genus_name,
                            " }\n\tfn get_species(&self) -> Box<dyn super::ObjectSpecies> { Box::new(",
                            family.get_name().as_str(),
                            "Species::",
                            object.get_name(),
                            ") }\n\tfn get_raw_object(&self) -> RawObject { self.raw_object.clone() }\n}\n\n"
                        ).as_str());
                    }
                }
            }
        },
        m if m.contains("members") => {
            write_string = format!(
                "{}{}{}", 
                "use super::MessageSpecies;\n\npub enum ", 
                family.get_name().as_str(), "Message{\n"
            );
            for messages in family.get_messages() {
                #[allow(unused_assignments)]
                let mut message_type = "";
                match family.get_messages().iter().position(|x| x.eq(&messages)) {
                    Some(0) => message_type = "Request",
                    Some(1) => message_type = "Response",
                    Some(2) => message_type = "Command",
                    Some(3) => message_type = "Status",
                    Some(4) => message_type = "Stream",
                    None => panic!("When writing file, message type error"),
                    Some(_) => panic!("When writing file, message type error")
                }
                let mut message_num = 0x00;
                for message in messages {
                    write_string.push_str(format!(
                        "\t{},\t\t\t//{} {}\n", 
                        message, 
                        message_type, 
                        message_num, 
                    ).as_str());
                    message_num += 1;
                }
            }
            write_string.push_str(format!(
                "}}\n\nimpl MessageSpecies for {}Message{{}} ", 
                family.get_name().as_str(),
            ).as_str());
            
        },
        &_ => panic!("Output path format error")
    }
    
    return (write_string, output_path);
}

fn write_file(args: (String, String)) {
    let write_string = args.0;
    let output_path = args.1;
    if !std::path::Path::new(&output_path).exists(){
        let file_create_return = std::fs::File::create(&output_path);
        match file_create_return {
            Err(error) => {
                println!("Failed to open file {} to write! Do you have correct permissions?",output_path);
                println!("Error code: {:?}",error);
                todo!();
            }
            _ => {}
        }
    }

    let mut possible_output_file = std::fs::OpenOptions::new().write(true).truncate(true).open(&output_path);
    match possible_output_file{
        Ok(ref mut output_file) =>{
            let write_result = output_file.write_all(write_string.as_bytes());
            match write_result{
                Err(write_error) => {
                    println!("Failed to write to file! Error: {:?}", write_error);
                    todo!();
                }
                _ => {}
            }
        },
        Err(error) =>{
            println!("Failed to open file {} to write! Do you have correct permissions?",output_path);
            println!("Error code: {:?}",error);
            todo!();
        }
    }
}
